import re, datetime, StringIO, gzip, os

class LogFormat:

	LogFormatCharacters = {
		'a' : ('([0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3})', ('ip_address',)),
		'b'	: ('([0-9-]+)', ('bytes_sent',)),
		'B'	: ('([0-9-]+)', ('bytes_sent',)),
		'h'	: ('([^ ]+)', ('remote_host',)),
		'm'	: ('([A-Za-z]+)', ('request_method',)),
		's'	: ('([0-9]{3})', ('request_status',)),
		'r'	: ('([^ ]+) (.+) HTTP/([0-9.]+)', ('request_method', 'uri', 'http_version')),
		'R'	: ('"([^"]*)"', ('referrer',)),
		't' : ('\[([0-9]{2})/([a-zA-Z]{3})/([0-9]{4}):([0-9]{2}):([0-9]{2}):([0-9]{2}) [0-9-]+\]',
				('date_day','date_month','date_year','time_hour','time_minute','time_second')),
		'u'	: ('([^ ]+)', ('uri',)),
		'U'	: ('"([^"]*)"', ('user_agent',)),
	}
	
	def __init__(self, log_format_string):
		self.log_format_string = log_format_string
		self.fields = []
		self.__convert_to_regex()

	def __convert_to_regex(self):
		regex_string = ['^']
		self.fields = []
		escaped = False

		for c in self.log_format_string:
			if escaped:
				escaped = False
				if LogFormat.LogFormatCharacters.has_key(c):
					(regex, fields) = LogFormat.LogFormatCharacters[c]
					regex_string.append(regex)
					self.fields.extend(fields)
				else:
					regex_string.append('%')
					regex_string.append(c)
			else:
				if c == '%': escaped = True
				else: regex_string.append(c)

		regex_string.append('$')
		self.regex = re.compile(''.join(regex_string))

class LogEntry:
	ip_regex = re.compile('^[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}$')
	months = ['Jan','Feb','Mar','Apr','May','Jun','Jul','Aug','Sep','Oct','Nov','Dec']
	
	def get_domain(self):
		if not self.remote_host: return self.ip_address
		if LogEntry.ip_regex.match(self.remote_host): return self.remote_host
		host_parts = self.remote_host.split('.')
		return '.'.join(host_parts[-2:])			
	
	def __init__(self, argdict, **kwargs):
		# Set some defaults
		self.ip_address = ''
		self.bytes_sent = 0
		self.remote_host = ''
		self.request_method = ''
		self.request_status = ''
		self.uri = ''
		self.http_version = ''
		self.referrer = ''
		self.user_agent = ''
		
		# Set values passed into constructor
		for (k,v) in argdict.items():
			setattr(self, k, v)

		for (k,v) in kwargs.items():
			setattr(self, k, v)

		# If bytes sent is -, convert to 0
		if self.bytes_sent and self.bytes_sent == '-':
			self.bytes_sent = 0
		else:
			self.bytes_sent = int(self.bytes_sent)

		# See if remote host is actually the IP
		if LogEntry.ip_regex.match(self.remote_host):
			self.ip_address = self.remote_host
			self.remote_host = ''

		# Get date
		try:
			self.entry_time = datetime.datetime(
								year = int(self.date_year),
								month = LogEntry.months.index(self.date_month)+1,
								day = int(self.date_day),
								hour = int(self.time_hour),
								minute = int(self.time_minute),
								second= int(self.time_second)
								)
		except:
			self.entry_time = None

class InvalidLogFormatError(Exception):
	pass

class LogParser:
	
	def __init__(self, log_filename=None, log_file_object=None, log_file_string=None):
		if log_filename and os.path.splitext(log_filename.lower()) == '.gz':
			self.filehandle = gzip.open(log_filename, 'r')
		elif log_filename:
			self.filehandle = file(log_filename)
		elif log_file_object:
			self.filehandle = log_file_object
		elif log_file_string:
			self.filehandle = StringIO.StringIO(log_file_string)
			
		self.formats = []
		self.last_line = ""

	def add_format(self, logformat):
		self.formats.append(logformat)
	
	def parse_line(self, line):
		line = line.strip()
		self.last_line = line
		for format in self.formats:
			m = format.regex.match(line)
			if m:
				linedata = {}
				for i in range(0, len(format.fields)):
					linedata[format.fields[i]] = m.group(i+1)
				return LogEntry(linedata)
		raise InvalidLogFormatError(line)
	
	def next(self):
		line = self.filehandle.next()
		if not line: return None
		return self.parse_line(line)
	
	def close(self):
		self.filehandle.close()
