import re

class ConfigFileError(Exception): pass

class ConfigParser:
	def __init__(self, config_filename):
		self.filename = (config_filename)
		self.parse_file()

	def parse_file(self):
		self.blocks = []
		currentblock = None
		linecount = 0
		f = file(self.filename)
		for line in f:
			line = line.strip()
			linecount += 1
			
			if not line or line.startswith('#'):
				continue
			elif line.startswith('[') and line.endswith(']'):
				currentblock = ConfigBlock(line[1:-1])
				self.blocks.append(currentblock)
			elif currentblock:
				directive = re.split('\s+', line)[0]
				value = line[len(directive):].strip()

				if directive in ('mysql_host', 'mysql_user', 'mysql_passwd', 'mysql_dbname', 'session_threshold'):
					if getattr(currentblock, directive, None):
						raise ConfigFileError('Directive %s is defined twice in block [%s], line %s' % (directive, currentblock.name, linecount))
					setattr(currentblock, directive, value)

				elif directive in ('max_age',):
					if getattr(currentblock, directive, None):
						raise ConfigFileError('Directive %s is defined twice in block [%s], line %s' % (directive, currentblock.name, linecount))
					setattr(currentblock, directive, int(value))
						
				elif directive == 'logfile':
					currentblock.add_logfile(value)

				elif directive == 'ignore_extension':
					currentblock.add_ignored_extension(value)

				elif directive == 'internal_referrer':
					currentblock.add_internal_referrer(value)

				elif directive == 'ignore_user_agent':
					currentblock.add_ignored_user_agent(value)

				elif directive == 'ignore_hostname':
					currentblock.add_ignored_hostname(value)

				elif directive == 'ignore_ip':
					currentblock.add_ignored_ip(value)

				elif directive == 'ignore_url':
					currentblock.add_ignored_url(value)

				elif directive == 'host_proxy':
					currentblock.add_host_proxy(value)

				elif directive == 'log_format':
					currentblock.add_log_format(value)
				
				# Accept for compatibility
				elif directive in ('connection_string', 'input_format', 'database_name', 'daily_logfile'):
					print "Ignoring \"%s\" directive for compatibility." % directive

				else:
					raise ConfigFileError('Unrecognized directive %s in block [%s], line %s' % (directive, currentblock.name, linecount))
			else:
				raise ConfigFileError('Directive %s not inside a block at line %s' % (directive, linecount))
		f.close()
						
					
					


class ConfigBlock:
	def __init__(self, blockname):
		self.name = blockname

		# Set defaults
		self.mysql_host = ''
		self.mysql_user = ''
		self.mysql_passwd = ''
		self.mysql_dbname = ''
		self.session_threshold = ''

		# Create lists
		self.logfiles = []
		self.ignored_extensions = []
		self.internal_referrers = []
		self.ignored_user_agents = []
		self.ignored_urls = []
		self.ignored_hostnames = []
		self.ignored_ips = []
		self.host_proxies = []
		self.log_formats = []
		self.max_age = None
	
	def add_logfile(self, filename):
		self.logfiles.append(filename)

	def add_ignored_extension(self, ext):
		self.ignored_extensions.append(ext)

	def add_internal_referrer(self, regex):
		self.internal_referrers.append(re.compile(regex, re.IGNORECASE))

	def add_ignored_user_agent(self, regex):
		self.ignored_user_agents.append(re.compile(regex, re.IGNORECASE))

	def add_ignored_url(self, regex):
		self.ignored_urls.append(re.compile(regex, re.IGNORECASE))

	def add_ignored_hostname(self, regex):
		self.ignored_hostnames.append(re.compile(regex, re.IGNORECASE))

	def add_ignored_ips(self, regex):
		self.ignored_ips.append(re.compile(regex, re.IGNORECASE))

	def add_host_proxy(self, regex):
		self.host_proxies.append(re.compile(regex, re.IGNORECASE))

	def add_log_format(self, log_format):
		self.log_formats.append(log_format)
