import MySQLdb
import socket
import os
import datetime

db = None

def set_database(host, user, passwd, dbname):
	global db
	if db: db.close()
	db = MySQLdb.connect(host, user, passwd, dbname)
	return db

def close_database():
	global db
	db.close()
	db = None

def optimize_tables():
	c = db.cursor()
	c.execute('''OPTIMIZE TABLE dns_resolution''')
	c.execute('''OPTIMIZE TABLE session''')
	c.execute('''OPTIMIZE TABLE pagehit''')

def clear_session_database():
	c = db.cursor()
	c.execute('''DROP TABLE IF EXISTS session''')
	c.execute('''DROP TABLE IF EXISTS page_hit''')

def clear_dns_database():
	c = db.cursor()
	c.execute('''DROP TABLE IF EXISTS dns_resolution''')

def create_database_structure():
	c = db.cursor()
	c.execute('''SHOW TABLES''')
	tables = [x[0] for x in c.fetchall()]

	if 'dns_resolution' not in tables:
		c.execute('''CREATE TABLE `dns_resolution` (
				`ip_address` VARCHAR( 255 ) NOT NULL,
				`host_name` VARCHAR( 255 ) NOT NULL,
				`primary_hostname` ENUM('Y','N') DEFAULT 'N' NOT NULL
				) ''')
		c.execute('''ALTER TABLE `dns_resolution` ADD INDEX `ip_address` (`ip_address`)''')
		c.execute('''ALTER TABLE `dns_resolution` ADD INDEX `host_name` (`host_name`)''')
	
	if 'session' not in tables:
		c.execute('''CREATE TABLE `session` (
				`session_id` INT UNSIGNED ZEROFILL NOT NULL AUTO_INCREMENT ,
				`ip_address` VARCHAR( 255 ) NOT NULL ,
				`host_name` VARCHAR( 255 ) NOT NULL ,
				`domain_name` VARCHAR(255) NOT NULL,
				`user_agent` VARCHAR( 255 ) NOT NULL ,
				`referrer` VARCHAR( 255 ) NOT NULL ,
				`first_hit_id` INT UNSIGNED ZEROFILL NOT NULL ,
				`first_hit` DATETIME NOT NULL ,
				`last_hit` DATETIME NOT NULL ,
				PRIMARY KEY ( `session_id` )
				)''')
		c.execute('''ALTER TABLE `session` ADD INDEX `ip_address` ( `ip_address` )''')
		c.execute('''ALTER TABLE `session` ADD INDEX `first_hit_id` ( `first_hit_id` )''')
		c.execute('''ALTER TABLE `session` ADD INDEX `host_name` ( `host_name` )''')
		c.execute('''ALTER TABLE `session` ADD INDEX `user_agent` ( `user_agent` )''')
		c.execute('''ALTER TABLE `session` ADD INDEX `referrer` ( `referrer` )''')
	
	if 'page_hit' not in tables:
		c.execute('''CREATE TABLE `page_hit` (
				`page_hit_id` INT UNSIGNED ZEROFILL NOT NULL AUTO_INCREMENT ,
				`session_id` INT UNSIGNED ZEROFILL NOT NULL ,
				`hit_time` DATETIME NOT NULL ,
				`http_method` VARCHAR( 15 ) NOT NULL ,
				`http_version` VARCHAR( 15 ) NOT NULL ,
				`http_status` INT NOT NULL ,
				`bytes_sent` INT NOT NULL ,
				`uri` VARCHAR( 255 ) NOT NULL ,
				PRIMARY KEY ( `page_hit_id` )
				)''')
		c.execute('''ALTER TABLE `page_hit` ADD INDEX `session_id` ( `session_id` )''')
		c.execute('''ALTER TABLE `page_hit` ADD INDEX `uri` ( `uri` )''')
	
	
class SessionLogger:

	def __init__(self, config):
		self.config = config

	def is_old_hit(self, log_entry):
		if self.config.max_age:
			age = datetime.datetime.now() - log_entry.entry_time
			return age.days > self.config.max_age
		return False
		

	def is_ignored_hit(self, log_entry):
		lower_uri = log_entry.uri.lower()
		
		# Check ignored extensions
		if self.config.ignored_extensions:
			uri_path = lower_uri.split('?')[0].split('#')[0]
			uri_ext = os.path.splitext(uri_path)[1][1:]
			for ext in self.config.ignored_extensions:
				if ext.lower() == uri_ext: return True

		# Check ignored user agents
		if log_entry.user_agent:
			for regex in self.config.ignored_user_agents:
				if regex.findall(log_entry.user_agent): return True

		# Check ignored URLS
		for regex in self.config.ignored_urls:
			if regex.findall(log_entry.uri): return True

		# Check ignored IPs
		self.resolve_hostname(log_entry)
		for regex in self.config.ignored_ips:
			if regex.fidnall(log_entry.ip_address): return True

		# Check ignored hostnames
		for regex in self.config.ignored_hostnames:
			if regex.findall(log_entry.remote_host): return True

		return False

	def is_duplicate_hit(self, session_id, log_entry):
		c = db.cursor()
		c.execute('''SELECT page_hit_id FROM page_hit WHERE session_id=%s AND hit_time=%s AND uri=%s''',
			   (session_id, log_entry.entry_time, log_entry.uri))
		return c.rowcount and True or False

	def resolve_hostname(self, log_entry):
		dr = DnsResolver()

		# Resolve DNS
		if not log_entry.remote_host:
			log_entry.remote_host = dr.gethostbyaddr(log_entry.ip_address)
		elif not log_entry.ip_address:
			log_entry.ip_address = dr.getaddrbyhost(log_entry.remote_host)
		
		# Check proxies setting - if it matches, replace with the match
		# This converts proxy servers like somerandomvalue.proxy.aol.com to .proxy.aol.com for ease of grouping
		for regex in self.config.host_proxies:
			m = regex.findall(log_entry.remote_host)
			if m:
				log_entry.remote_host = m[0]
				break
	
	def find_session_id(self, log_entry):
		c = db.cursor()
		
		self.resolve_hostname(log_entry)
		
		# See if this is from a new referrer
		new_referrer = True
		log_entry.new_session = False
		if not log_entry.referrer or log_entry.referrer == '-':
			new_referrer = False

		if new_referrer:
			for regex in self.config.internal_referrers:	
				if regex.findall(log_entry.referrer):
					new_referrer = False
					break
				
		# See if there is a matching recent session (as long as it's not from a new referrer
		if not new_referrer:
			if not self.config.session_threshold:
				self.config.session_threshold = "1 HOUR"
			c.execute('''SELECT session_id FROM session
				     WHERE (ip_address=%%s OR host_name=%%s)
				     AND user_agent=%%s
				     AND first_hit <= %%s
				     AND DATE_ADD(last_hit, INTERVAL %s) >= %%s''' % self.config.session_threshold,
				     (log_entry.ip_address, log_entry.remote_host, log_entry.user_agent, log_entry.entry_time, log_entry.entry_time))
			if c.rowcount:
				return c.fetchone()[0]
		else:
			# Make sure this isn't a duplicate session
			c.execute('''SELECT session_id FROM session
				     WHERE (ip_address=%s OR host_name=%s)
				     AND user_agent=%s
				     AND first_hit=%s''',
				     (log_entry.ip_address, log_entry.remote_host, log_entry.user_agent, log_entry.entry_time))
			if c.rowcount:
				return c.fetchone()[0]
	
		# Create a new session
		c.execute('''INSERT INTO session SET ip_address=%s, host_name=%s, domain_name=%s, user_agent=%s, referrer=%s, first_hit=%s, last_hit=%s''',
				(log_entry.ip_address, log_entry.remote_host, log_entry.get_domain(), log_entry.user_agent, log_entry.referrer, log_entry.entry_time,
				 log_entry.entry_time))
		log_entry.new_session = True
		try:
			return c.insert_id()
		except:
			return c.lastrowid

	def log_hit(self, session_id, log_entry):
		if not self.is_duplicate_hit(session_id, log_entry):
			c = db.cursor()
			
			c.execute('''INSERT INTO page_hit SET session_id=%s, hit_time=%s, http_method=%s, http_version=%s, http_status=%s, bytes_sent=%s, uri=%s''',
					(session_id, log_entry.entry_time, log_entry.request_method, log_entry.http_version, log_entry.request_status,
					 log_entry.bytes_sent, log_entry.uri))
			try:
				page_hit_id = c.insert_id()
			except:
				page_hit_id = c.lastrowid
					 
			c.execute('''UPDATE session SET last_hit=%s WHERE session_id=%s AND last_hit < %s''',
					(log_entry.entry_time, session_id, log_entry.entry_time))
			if log_entry.new_session:
				c.execute('''UPDATE session SET first_hit_id=%s WHERE session_id=%s''', (page_hit_id, session_id))



class DnsResolver:
	host_names = {}
	ip_addresses = {}

	def gethostbyaddr(self, ip_address):
		c = db.cursor()
		
		# Dictionary in memory
		if DnsResolver.host_names.has_key(ip_address):
			return DnsResolver.host_names[ip_address]
		else:
			# Database cache
			c.execute('''SELECT host_name FROM dns_resolution WHERE ip_address=%s AND primary_hostname="Y"''', (ip_address,))
			if c.rowcount:
				host_name = c.fetchone()[0]
				DnsResolver.host_names[ip_address] = host_name
				return host_name
			
			# DNS lookup
			try:
				socket.setdefaulttimeout(60)
				host_name = socket.gethostbyaddr(ip_address)[0]
			except:
				host_name = ''

			DnsResolver.host_names[ip_address] = host_name
			c.execute('''UPDATE dns_resolution SET primary_hostname="Y" WHERE host_name=%s AND ip_address=%s''', (host_name, ip_address))
			if not c.rowcount:
				c.execute('''INSERT INTO dns_resolution SET host_name=%s, ip_address=%s, primary_hostname="Y"''', (host_name, ip_address))
			return host_name
				
	def getaddrbyhost(self, host_name):
		c = db.cursor()

		# Dictionary in memory
		if DnsResolver.ip_addresses.has_key(host_name):
			return DnsResolver.ip_addresses[host_name]
		else:
			# Database cache
			c.execute('''SELECT ip_address FROM dns_resolution WHERE host_name=%s''', (host_name,))
			if c.rowcount:
				ip_address = c.fetchone()[0]
				DnsResolver.ip_addresses[host_name] = ip_address
				return ip_address

			# DNS lookup
			try:
				socket.setdefaulttimeout(60)
				ip_address = socket.gethostbyname(host_name)
			except:
				ip_address = ''

			DnsResolver.ip_addresses[host_name] = ip_address
			c.execute('''INSERT INTO dns_resolution SET host_name=%s, ip_address=%s''', (host_name, ip_address))
			return ip_address
